/*****************************************************************************
 * $Id: game_base_chase_plane.cc,v 1.14 2006/07/28 18:17:46 devel Exp $
 * @file
 * @author	Jim E. Brooks  http://www.jimbrooks.org
 * @brief	ChasePlane classes.
 * @remarks
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE (GPL).
 *****************************************************************************/

#define GAME_BASE_CHASE_PLANE_CC 1
#include "game_base_common.hh"
#include "game_base_chase_plane.hh"

////////////////////////////////////////////////////////////////////////////////
////////////////////////////  ChasePlaneCraft  /////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

MatrixExt ChasePlaneCraft::mXmSaved;
fp  ChasePlaneCraft::mDistSaved;

/*****************************************************************************
 * ChasePlane.
 *****************************************************************************/
ChasePlaneCraft::ChasePlaneCraft( const Dyna& owner )
: ChasePlane(owner)
{
	Reset();
}

/*****************************************************************************
 * Reset chase-plane view.
 *****************************************************************************/
void
ChasePlaneCraft::Reset( void )
{
	mDist = WORLD(500.0);
	IdentityMatrix( mXm );
}

/*****************************************************************************
 * Transform the chase plane to track its target.
 * Changes the eye's matrix.
 * Called by Eye::Animate().
 * @returns True if redisplay needed.
 *****************************************************************************/
int
ChasePlaneCraft::Chase( void )
{
	// Zoom chase-plane's matrix away from the target (translate from origin).
	// Then copy the chase-plane's matrix into the eye's matrix.
	ZoomOutFromTarget();
	gEng.mEye.LoadMatrix( mXm );
	gEng.mEye.Translate( AXIS_X, 0, true );		// to trigger update
	return true;
}

/*****************************************************************************
 * - Copy target's origin into chase-plane's origin.
 * - Translate chase plane in -Z direction (translate from origin)
 *   to move chase plane backwards relative to target.
 *****************************************************************************/
void
ChasePlaneCraft::ZoomOutFromTarget( void )
{
	mXm[Ox] = mOwner.mCm[Ox];		// assigning with single-precision values is ok
	mXm[Oy] = mOwner.mCm[Oy];		// as rotation elements, not origin elements,
	mXm[Oz] = mOwner.mCm[Oz];		// need extended-precision
	Translate( AXIS_Z, mDist );
}

/*****************************************************************************
 * For keyboard control.
 * Incrementally rotate chase-plane by a default degree.
 * @param	axis
 * @param	dir
 *			{1,-1}
 *****************************************************************************/
void
ChasePlaneCraft::IncRotate( uint axis, int dir )
{
	Save();
		const fp deg = 2.0;
		Rotate( axis, dir > 0 ? deg : -deg );
	RestoreIfInvalid();
}

/*****************************************************************************
 * For keyboard control.
 * Increment chase plane distance (zoom in/out) by a default amount.
 * @param	dir
 *			{1,-1}
 *****************************************************************************/
void
ChasePlaneCraft::IncDistance( int dir )
{
	Save();
		mDist += dir * WORLD(10.0);
		mDist = Range( mDist, (fp)WORLD(200.0), (fp)WORLD(2000.0) ); // stay in bounds
	RestoreIfInvalid();
}

/*****************************************************************************
 * For keyboard control.
 * Set chase plane distance (by assignment).
 *****************************************************************************/
void
ChasePlaneCraft::SetDistance( fp dist )
{
	mDist = dist;
}

/*****************************************************************************
 * @return Chase-plane distance from its target.
 *****************************************************************************/
fp
ChasePlaneCraft::GetDistance( void )
{
	return mDist;
}

/*****************************************************************************
 * Save state in case an operation makes it invalid (illegal view).
 *****************************************************************************/
void
ChasePlaneCraft::Save( void )
{
	CopyMatrix( mXmSaved, mXm );
	mDistSaved = mDist;
}

/*****************************************************************************
 * Restore state if an operation has made it invalid (illegal view).
 *****************************************************************************/
void
ChasePlaneCraft::RestoreIfInvalid( void )
{
	// Test zooming out from target.
	// Restore saved matrix if origin was zoomed below ground.
	ZoomOutFromTarget();
	if ( gWorld.IfBelowGround( mXm ) )
	{
		CopyMatrix( mXm, mXmSaved );
		mDist = mDistSaved;
	}
}

/*****************************************************************************
 * Rotate chase-plane.
 *****************************************************************************/
void
ChasePlaneCraft::Rotate( uint axis, fp deg )
{
	RotateMatrixFixed( axis, (fpx)deg, mXm );
}

/*****************************************************************************
 * Translate chase-plane.
 *****************************************************************************/
void
ChasePlaneCraft::Translate( uint axis, fp inc )
{
	TranslateMatrixFixed( axis, (fpx)inc, mXm );
}
